//----------------------- INFO ----------------------
// REGELTRAFO
// CPU: PIC8F2685
//
//
//----------------------- IDE -----------------------
// MPLAB IDE: 8.84
// Compiler: C18 V3.34
//
//----------------- COMPILER OPTIONS ----------------
// none
//

//--------------------- INCLUDES --------------------
#include	<p18cxxx.h>
#include	<delays.h>
#include	<math.h>
#include	"Regeltrafo.h"


//------------------- CONFIG BITS -------------------
#pragma config OSC = IRCIO67		// Internal oscillator block, CLKO function on RA6, port function on RA7
#pragma config FCMEN = OFF			// Fail-Safe Clock Monitor disabled
#pragma config IESO = OFF			// Internal/External Switchover mode disabled
#pragma config PWRT = ON			// Power-up Timer enabled
#pragma config BOREN = BOHW			// Brown-out Reset enabled in hardware only (SBOREN is disabled)
#pragma config BORV = 0				// Brown-out Reset Voltage set to maximum (4.5V)
#pragma config WDT = OFF			// Watchdog Timer disabled (control is placed on the SWDTEN bit)
#pragma config WDTPS = 1			// Watchdog Timer Postscale 1:1
#pragma config PBADEN = OFF			// PORTB<4:0> pins are configured as digital I/O on Reset
#pragma config LPT1OSC = OFF		// Timer1 configured for higher power operation
#pragma config MCLRE = ON			// MCLR pin enabled; RE3 input pin disabled
#pragma config STVREN = ON			// Stack full/underflow will cause Reset
#pragma config LVP = OFF			// Single-Supply ICSP disabled
#pragma config BBSIZ = 1024			// 1K words (2K bytes) Boot Block
#pragma config XINST = OFF			// Instruction set extension and Indexed Addressing mode disabled (Legacy mode)
#pragma config CP0 = OFF			// Block 0 not code-protected  
#pragma config CP1 = OFF			// Block 1 not code-protected  
#pragma config CP2 = OFF			// Block 1 not code-protected  
#pragma config CP3 = OFF			// Block 1 not code-protected  
#pragma config CP4 = OFF			// Block 1 not code-protected  
#pragma config CP5 = OFF			// Block 1 not code-protected  
#pragma config CPB = OFF			// Boot block (000000-0001FFh) is not code-protected
#pragma config CPD = OFF			// Data EEPROM is not code-protected
#pragma config WRT0 = OFF			// Block 0 not write-protected
#pragma config WRT1 = OFF			// Block 1 not write-protected  
#pragma config WRT2 = OFF			// Block 1 not write-protected  
#pragma config WRT3 = OFF			// Block 1 not write-protected  
#pragma config WRT4 = OFF			// Block 1 not write-protected  
#pragma config WRT5 = OFF			// Block 1 not write-protected  
#pragma config WRTC = OFF			// Configuration registers (300000-3000FFh) are not write-protected  
#pragma config WRTB = OFF			// Boot block (000000-0001FFh) is not write-protected  
#pragma config WRTD = OFF			// Data EEPROM is not write-protected 
#pragma config EBTR0 = OFF			// Block 0 not protected from table reads executed in other blocks
#pragma config EBTR1 = OFF			// Block 1 not protected from table reads executed in other blocks  
#pragma config EBTR2 = OFF			// Block 1 not protected from table reads executed in other blocks  
#pragma config EBTR3 = OFF			// Block 1 not protected from table reads executed in other blocks  
#pragma config EBTR4 = OFF			// Block 1 not protected from table reads executed in other blocks  
#pragma config EBTR5 = OFF			// Block 1 not protected from table reads executed in other blocks  
#pragma config EBTRB = OFF			// Boot block (000000-0001FFh) is not protected from table reads executed in other blocks


//----------------- GLOBAL VARIABLES ----------------
unsigned char	trigger = 0x0;
unsigned char	display_fuse = 0x1;
unsigned char	display_power = 0x1;
unsigned char	fuse_mode = FUSE_WATT;
unsigned char	power_mode = POWER_REAL;
unsigned long  	volt = 0x0;
unsigned long  	amps = 0x0;
unsigned long  	power = 0x0;
unsigned int	fuse = 10;
unsigned int	ADCcount = 0x0;
unsigned int	volts_zero = 403;
unsigned int	amps_zero = 403;

//                      0123456789012345
static char line1[]  = "U:    V  I:    A";
static char line2[]  = "P:     W F:     ";
static char pline1[] = "Please turn the ";
static char pline2[] = "I/O switch OFF  ";
static char wline1[] = "REGELTRAFO  V1.0";
static char wline2[] = "dw 271012       ";



//-----------------     M A I N      ----------------

void main (void) {

	unsigned long  	i;
	double			pwr_real;
	double			pwr_apparent;
	signed char		j;
	unsigned int	adc;
	unsigned long	volts_new = 0x0;
	unsigned long	amps_new = 0x0;
	unsigned int	u;
	unsigned int	timer3;
	unsigned char	key_onoff = 0x1;
	unsigned int	key_plus = 0x0;
	unsigned int	key_minus = 0x0;
	unsigned char	fuse_blink = 0x0;
	char			lcdtext[10];


	// Initialize CPU and peripherals

	CPU_Init();													// Initialize CPU
	LCD_Init();													// Initialize LCD


	// Print MOTD

	LCD_WriteString (wline1);									// print welcome line 1
	LCD_SetCursor(0,1);
	LCD_WriteString (wline2);									// print welcome line 2
	LCD_SetCursor(0,0);
	for (i=0; i<3; i++) {										// wait 3 seconds
		Delay10KTCYx(200);
	}


	// Check for ON/OFF key

	if (_KEY_ONOFF == 0x0) {									// if ON/OFF switch is set to 'ON'
		LCD_WriteString (pline1);								// ask user to switch it OFF
		LCD_SetCursor(0,1);
		LCD_WriteString (pline2);
		while (_KEY_ONOFF == 0x0);								// wait until switch is OFF
		LCD_SetCursor(0,0);
	}


	// Print header lines

	LCD_WriteString (line1);									// print header line 1
	LCD_SetCursor(0,1);
	LCD_WriteString (line2);									// print header line 2
	LCD_SetCursor(7,0);
	LCD_senddata(LCD_DATA, 0x0);								// print '~'

	TMR0_Init();												// Start Timer0


	// Main loop

	while (1) {

		// SAMPLE VOLTS

		ADCON0 = 0b00000001;													// select channel AN0 to sample volts
		ADCON0bits.GO_DONE = 1;													// start sampling
		while (ADCON0bits.GO_DONE == 1) {};										// wait for ADC to finish
		adc = ((unsigned int)(ADRESH)<<8)+ADRESL;								// get ADC value
		volts_new += adc;														// sum up value to calculate zero line
		u = 0x0;
		if (adc > volts_zero) {													// rectify value
			volt += (unsigned long)(adc-volts_zero)*(adc-volts_zero);			// square value above zero line
			u = adc-volts_zero;													// save volts to calculate power
		} else if (adc < volts_zero) {
			volt += (unsigned long)(volts_zero-adc)*(volts_zero-adc);			// square value below zero line
			u = volts_zero-adc;													// save volts to calculate power
		}


		// SAMPLE AMPS

		ADCON0 = 0b00000101;													// select channel AN1 to sample amps
		ADCON0bits.GO_DONE = 1;													// start sampling
		while (ADCON0bits.GO_DONE == 1) {};										// wait for ADC to finish
		adc = ((unsigned int)(ADRESH)<<8)+ADRESL;								// get ADC value
		amps_new += adc;														// sum up value to calculate zero line
		if (adc > (amps_zero+2)) {												// rectify value
			amps += (unsigned long)(adc-amps_zero)*(adc-amps_zero);				// square value above zero line
			power += (unsigned long)(adc-amps_zero)*u;							// calculate power
		} else if (adc < (amps_zero-2)) {
			amps += (unsigned long)(amps_zero-adc)*(amps_zero-adc);				// square value below zero line
			power += (unsigned long)(amps_zero-adc)*u;							// calculate power
		}

		ADCcount++;



		// CHECK FOR KEYS PRESSED //

		timer3 = TMR3L;
		timer3 += (unsigned int)TMR3H<<8;										// get timer3 value

		if (_KEY_ONOFF != key_onoff) {											// Check for movement of ON/OFF key
			if (timer3 == 0x0) {												// continue if key did not bounce
				if (key_onoff == 0x0) {											// key was released
					_RELAIS = 0x0;												// turn relais off
					fuse_blink = 0x0;											// stop fuse blinking
					display_fuse = 0x1;
				} else {														// key was pressed
					_RELAIS = 0x1;												// turn relais on
				}
			}
			TMR3_Init(50);														// suppress key bouncing
			key_onoff = 1-key_onoff;											// save key state
		}

		if ((_KEY_PLUS == 0x1) && (key_plus > 0x0)) {							// Key PLUS was released
			key_plus = 0x0;
			TMR3_Init(250);														// suppress key bouncing
		}
		if (timer3 == 0x0 && _KEY_PLUS == 0x0) {								// Check for key PLUS
			TMR3_Init(250);														// suppress key bouncing
			if (key_onoff == 0x0) {												// RUN mode
				if (key_plus == 0x0) {											// increase power mode only once per key pressing
					power_mode++;												// display next power value
					if (power_mode == POWER_LASTMODE) {
						power_mode = 0x0;
					}
				}
			} else {															// SETUP mode
				if ((key_plus%2) == 0x0) {										// key_plus is the counter for TMR3 overflows
					if (key_plus < 10) {										// slow mode (increase fuse by one)
						fuse++;
						display_fuse = 0x1;
					} else if (key_plus == 10) {								// entering fast mode (round value)
						fuse = fuse - (fuse%10) +10;							// round to next decade
						display_fuse = 0x1;
					} else {													// fast mode (increase fuse by 10)
						fuse += 10;
						display_fuse = 0x1;
					}
					if (fuse_mode == FUSE_WATT) {
						if (fuse > _FUSE_MAX_WATT) {							// maximum Watts reached
							fuse_mode = FUSE_AMPS;								// switch to Amps
							fuse = 1;
							key_plus = 0x1;
						}
					}
					if (fuse_mode == FUSE_AMPS) {
						if (fuse > _FUSE_MAX_AMPS) {							// maximum Amps reached
							fuse_mode = FUSE_WATT;								// switch to Watt
							fuse = 1;
							key_plus = 0x1;
						}
					}
				}
			}
			key_plus++;
		}


		if ((_KEY_MINUS == 0x1) && (key_minus > 0x0)) {							// Key MINUS was released
			key_minus = 0x0;
			TMR3_Init(250);														// suppress key bouncing
		}
		if (timer3 == 0x0 && _KEY_MINUS == 0x0) {								// Check for key MINUS
			TMR3_Init(250);														// suppress key bouncing
			if (key_onoff == 0x0) {												// RUN mode
				if (key_minus == 0x0) {											// decrease power mode only once per key pressing
					if (power_mode == 0x0) {
						power_mode = POWER_LASTMODE;
					}
					power_mode--;												// display next power value
				}
			} else {															// SETUP mode
				if ((key_minus%2) == 0x0) {											// key_minus is the counter for TMR3 overflows
					if (key_minus < 10) {											// slow mode (decrease fuse by one)
						fuse--;
						display_fuse = 0x1;
					} else if (key_minus == 10) {									// entering fast mode (round value)
						fuse = fuse - (fuse%10);									// round to next decade
						display_fuse = 0x1;
					} else {														// fast mode (decrease fuse by 10)
						fuse -= 10;
						display_fuse = 0x1;
					}
					if (fuse_mode == FUSE_WATT) {
						if (fuse == 0x0) {											// minimum Watts reached
							fuse_mode = FUSE_AMPS;									// switch to Amps
							fuse = _FUSE_MAX_AMPS;
							key_minus = 0x1;
						}
					}
					if (fuse_mode == FUSE_AMPS) {
						if (fuse == 0x0) {											// minimum Amps reached
							fuse_mode = FUSE_WATT;									// switch to Watt
							fuse = _FUSE_MAX_WATT;
							key_minus = 0x1;
						}
					}
				}
			}
			key_minus++;
		}



		// PRINT FUSE VALUE IF NECESSARY

		if (display_fuse > 0x0) {
			LCD_SetCursor(11,1);
			for (j=0; j<10; j++){
				lcdtext[j] = 0x0;
			}
			if (fuse_blink == 0x1) {					// if the fuse value shall blink,
				for (j=0; j<5; j++){
					lcdtext[j] = ' ';					// clear the fuse field
				}					
			} else {
				i = fuse;
				switch (fuse_mode) {
					case FUSE_WATT:	lcdtext[4]='W';
									lcdtext[3]='0'+i%10;
									i/=10;
									lcdtext[2]='0'+i%10;
									i/=10;
									lcdtext[1]='0'+i%10;
									lcdtext[0]=' ';
									if (lcdtext[1] == '0' && lcdtext[2] == '0') {
										lcdtext[2] = ' ';
									}
									if (lcdtext[1] == '0') {
										lcdtext[1] = ' ';
									}
									break;
					case FUSE_AMPS:	lcdtext[4]='A';
									lcdtext[3]='0'+i%10;
									i/=10;
									lcdtext[2]='0'+i%10;
									i/=10;
									lcdtext[1]='.';
									lcdtext[0]='0'+i%10;
									break;
				}
			}
			LCD_WriteString (lcdtext);
			display_fuse = 0x0;
		}


		// CHECK FOR TRIGGER
		
		if (trigger > 0x0) {

			if (fuse_blink > 0x0) {								// if the fuse value shall blink,
				fuse_blink++;									// toggle the fuse display
				if (fuse_blink > 0x2) {
					fuse_blink = 0x1;
				}
				display_fuse = 0x1;
			}


			// PRINT VOLTS

			i = (sqrt(volt/ADCcount))*0.9;						// VOLTS
			LCD_SetCursor(3,0);
			for (j=0; j<10; j++){
				lcdtext[j] = 0x0;
			}
			for (j=2; j>-1; j--){
				lcdtext[j]='0'+i%10;
				i/=10;
			}
			if (lcdtext[0] == '0' && lcdtext[1] == '0') {
				lcdtext[1] = ' ';
			}
			if (lcdtext[0] == '0') {
				lcdtext[0] = ' ';
			}
			LCD_WriteString (lcdtext);


			// PRINT AMPS

			i = sqrt(amps/ADCcount)*0.8;						// AMPS
			LCD_SetCursor(11,0);
			for (j=0; j<10; j++){
				lcdtext[j] = 0x0;
			}
			lcdtext[3]='0'+i%10;
			i/=10;
			lcdtext[2]='0'+i%10;
			i/=10;
			lcdtext[1]='.';
			lcdtext[0]='0'+i%10;
			LCD_WriteString (lcdtext);


			// PRINT POWER

			LCD_SetCursor(0,1);
			for (j=0; j<10; j++){
				lcdtext[j] = 0x0;
			}
			lcdtext[1]=':';
			switch (power_mode) {
				case POWER_REAL:		i = (power/ADCcount)*0.072;											// Wirkleistung
										lcdtext[0]='P';
										lcdtext[7]='W';
										lcdtext[8]=' ';
										break;
				case POWER_APPARENT:	i = (sqrt((volt/ADCcount)*(amps/ADCcount)))*0.072;					// Scheinleistung
										lcdtext[0]='S';
										lcdtext[7]=0x1;		// character 'V'
										lcdtext[8]=0x2;		// character 'A'
										break;
				case POWER_COMPLEX:		pwr_real = (power/ADCcount)*0.072;									// Wirkleistung
										pwr_apparent = (sqrt((volt/ADCcount)*(amps/ADCcount)))*0.072;		// Scheinleistung
										i = sqrt(pwr_apparent*pwr_apparent - pwr_real*pwr_real);			// Blindleistung
										lcdtext[0]='Q';
										lcdtext[7]=0x3;		// character 'v'
										lcdtext[8]=0x4;		// character 'r'
										break;
			}
			lcdtext[6]='0'+i%10;
			i/=10;
			lcdtext[5]='.';
			lcdtext[4]='0'+i%10;
			i/=10;
			lcdtext[3]='0'+i%10;
			i/=10;
			lcdtext[2]='0'+i%10;
			if (lcdtext[3] == '0' && lcdtext[2] == '0') {
				lcdtext[3] = ' ';
			}
			if (lcdtext[2] == '0') {
				lcdtext[2] = ' ';
			}
			LCD_WriteString (lcdtext);


			// CHECK IF FUSE HAS TO BLOW
		
			if ((key_onoff == 0x0) && (fuse_blink == 0x0)) {					// check for RUN mode
				if (fuse_mode == FUSE_WATT) {
					if (((power/ADCcount)*0.072) >= (fuse*10)) {				// Real power limit reached?
						_RELAIS = 0x0;											// turn relais off
						fuse_blink = 0x1;										// blink fuse value to indicate that fuse is blown
					}
				}
				if (fuse_mode == FUSE_AMPS) {
					if ((sqrt(amps/ADCcount)*0.8) >= fuse) {					// Amp value limit reached?
						_RELAIS = 0x0;											// turn relais off
						fuse_blink = 0x1;										// blink fuse value to indicate that fuse is blown
					}
				}
			}
		

			// RESET VARIABLES

			volts_zero = volts_new / ADCcount;
			amps_zero = amps_new / ADCcount;
			ADCcount = 0x0;
			volt = 0x0;
			amps = 0x0;
			volts_new = 0x0;
			amps_new = 0x0;
			power = 0x0;
			trigger = 0x0;
		}
	}
}

